from itertools import combinations
import logging
import pytest
from aperturedb.EntityDataCSV import EntityDataCSV
from aperturedb.ParallelLoader import ParallelLoader
from aperturedb.ParallelQuery import ParallelQuery
from aperturedb.ParallelQuerySet import ParallelQuerySet
from aperturedb.Subscriptable import Subscriptable
from aperturedb.Connector import Connector
import math

logger = logging.getLogger(__name__)


class EntityWithResponseDataCSV(EntityDataCSV):
    def __init__(self, filename, requests, responses):
        super().__init__(filename)
        self.requests = requests
        self.responses = responses

    def response_handler(self, request, input_blob, response, output_blob):
        self.requests.append(request)
        self.responses.append(response)


class GeneratorWithNCommands(Subscriptable):
    def __init__(self, commands_per_query=1, blobs_per_query=0, elements=100) -> None:
        super().__init__()
        self.commands_per_query = commands_per_query
        self.blobs_per_query = blobs_per_query
        self.elements = elements

    def __len__(self):
        return self.elements

    def getitem(self, subscript):
        query = []
        blobs = []
        for i in range(self.commands_per_query):
            query.append({
                "FindEntity": {

                }
            })
        buff = None
        with open(__file__, "rb") as ins:
            buff = ins.read()

        for i in range(self.blobs_per_query):
            blobs.append(buff)

        return query, blobs


class QGPersons(Subscriptable):
    def __init__(self, requests, responses, cpq, response_blobs) -> None:
        super().__init__()
        self.requests = requests
        self.responses = responses
        self.response_blobs = response_blobs
        self.cpq = cpq

    def __len__(self):
        return math.ceil(10 / self.cpq)

    def getitem(self, subscript):
        query = []
        for i in range(self.cpq):
            query.append(
                {
                    "FindEntity": {
                        "with_class": "Person",
                        "results": {
                            "all_properties": True
                        },
                        "constraints": {
                            "age": [">=", (subscript * self.cpq + i) * 10, "<", (subscript * self.cpq + 1 + i) * 10]
                        }
                    }
                }
            )
            if subscript * self.cpq + 1 + i == 10:
                break
        return query, []

    def response_handler(self, request, input_blob, response, output_blob):
        self.requests.append(request)
        self.responses.append(response)
        if output_blob is not None and len(output_blob) > 0:
            self.response_blobs.append(output_blob)


class QGPersonsBadHandler(QGPersons):
    def response_handler(self, request, input_blob):
        self.requests.append(request)
        self.responses.append(response)
        if output_blob is not None and len(output_blob) > 0:
            self.response_blobs.append(output_blob)


class QGPersonsIndex(QGPersons):
    def response_handler(self, request, input_blob, response, output_blob, index):
        self.requests[index] = request
        self.responses[index] = response
        if output_blob is not None and len(output_blob) > 0:
            self.response_blobs[index] = output_blob


class QGImages(QGPersons):
    def getitem(self, subscript):
        query = []
        for i in range(self.cpq):
            q = {
                "FindImage": {
                    "blobs": True
                }
            }
            query.append(q)
        return query, []

# mimics a Entity connected to an Image


class QGSetPersonAndImages(Subscriptable):
    def __init__(self, requests, responses, max_commands, request_blobs) -> None:
        super().__init__()
        self.requests = requests
        self.responses = responses
        self.request_blobs = request_blobs
        self.max_commands = max_commands
        self.commands_per_query = [1, 1]
        self.blobs_per_query = [0, 1]

    def __len__(self):
        return self.max_commands

    def getitem(self, subscript):
        query_set = []
        entityquery = {
            "AddEntity": {
                "with_class": "Person",
                "properties": {
                    "age": (subscript + 20)
                },
                "constraint": {
                    "age": ["==", (subscript + 20)]
                }
            }
        }
        # add image if entity doesn't exist.
        image_constraint = {"results": {0: {"status": ["!=", 2]}}}

        imagequery = {
            "AddImage": {
                "properties": {
                    "type": "portrait",
                    "age": (subscript + 20)
                }
            }
        }
        query_set.append(entityquery)
        query_set.append([image_constraint, imagequery])

        entity_blobs = []
        image_blobs = [subscript]
        set_blobs = [entity_blobs, image_blobs]
        return [query_set], [set_blobs]

    def response_handler(self, set_id, request, input_blob, response, output_blob):
        if not set_id in self.requests:
            self.requests[set_id] = []
            self.responses[set_id] = []
            self.request_blobs[set_id] = []
        self.requests[set_id].append(request)
        self.responses[set_id].append(response)
        # we attach 1 input "blob",so take it out of the list.
        if input_blob is not None and len(input_blob) > 0:
            self.request_blobs[set_id].append(input_blob[0])


# Fake DB query.
def mock_query(self, request, blobs):
    self.response = 0
    return [], []

# Fake DB query. Depending in the transaction, return a response increasing number of images.
# Also add a item (int) to the blobs list for each image returned (to simulate the image data)
# The number of blobs returned is validated at the end.


def query_mocker_factory(response_blobs_count):
    def mock_query(self, request, blobs):
        self.response = 0
        mock_responses = [{
            "FindImage": {
                "returned": (i + 1),
                "status": 0
            }
        } for i, c in enumerate(request)]
        mock_blobs = [i for i in range(
            int(response_blobs_count * len(request)))]
        return mock_responses, mock_blobs
    return mock_query


def set_query_mocker_factory(response_blobs_count):
    # we return that entity "exists" for odd numbered
    # requests, this allows us to verify the results
    #  properly filter passed in blobs.
    # also return 0 for addimage.
    def mock_query(self, request, blobs):
        self.response = 0
        cmd_is_image = "AddImage" in request[0].keys()
        cmd = "AddImage" if cmd_is_image else "AddEntity"

        mock_responses = [{
            cmd: {
                "returned": (i + 1),
                "status": 0 if i % 2 == 0 or cmd_is_image else 2
            }
        } for i, c in enumerate(request)]
        mock_blobs = [i for i in range(
            int(response_blobs_count * len(request)))]
        return mock_responses, mock_blobs if cmd_is_image else []
    return mock_query


class TestResponseHandler():
    def cleanDB(self):
        # we no longer clean db since we insert no data, using mock instead.
        logger.debug(f"Cleaning existing data")
        self.requests = []
        self.responses = []

    @pytest.mark.parametrize("records_count", [(1), (99), (100)])
    def test_response_handler(self, records_count, db, utils, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s,
                            r, b: mock_query(s, r, b))
        self.cleanDB()
        logger.warning(f"{records_count}")
        persons = EntityWithResponseDataCSV(
            "./input/persons.adb.csv", self.requests, self.responses)
        persons = persons[:records_count]
        loader = ParallelLoader(db)
        loader.ingest(persons, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert loader.error_counter == 0

        assert len(persons) == records_count
        logger.info(self.responses)
        assert len(self.responses) == records_count
        logger.info(self.requests)
        assert len(self.requests) == records_count

    @pytest.mark.parametrize("commands_blobs", combinations(range(1, 4), 2))
    def test_varying_commands_blobs(self, db, commands_blobs, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s,
                            r, b: mock_query(s, r, b))
        i, j  = commands_blobs
        generator = GeneratorWithNCommands(i, j)
        querier = ParallelQuery(db, dry_run=True)
        querier.query(generator, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert querier.commands_per_query == i
        assert querier.blobs_per_query == j

    @pytest.mark.parametrize("cpq", range(1, 11))
    def test_varying_response(self, db, utils, cpq):
        self.cleanDB()
        persons = EntityWithResponseDataCSV(
            "./input/persons.adb.csv", self.requests, self.responses)
        loader = ParallelLoader(db)
        loader.ingest(persons, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert loader.error_counter == 0

        dist = persons.df.groupby(persons.df.age // 10 * 10).count()
        logger.debug(dist["age"])

        self.requests = []
        self.responses = []
        self.response_blobs = []
        generator = QGPersons(self.requests, self.responses,
                              cpq, self.response_blobs)
        querier = ParallelQuery(db)
        querier.query(generator, batchsize=100,
                      numthreads=1,
                      stats=True)
        dist_by_ages = {}
        for req, resp in zip(self.requests, self.responses):
            logger.debug(req)
            assert len(req) in [cpq, 10 % cpq]
            for rq, rr in zip(req, resp):
                for key in rq:
                    age_group = rq[key]["constraints"]["age"][1]
                    count = rr[key]["returned"]
                    dist_by_ages[age_group] = count
        logger.debug(dist_by_ages)
        for k in dist_by_ages:
            assert dist_by_ages[k] == dist["age"][k]

    def test_response_blobs(self, db, utils, monkeypatch):
        # 5.5 makes blob result uneven number compared to query count
        monkeypatch.setattr(Connector, "query", lambda s, r,
                            b: query_mocker_factory(5.5)(s, r, b))
        self.cleanDB()
        self.response_blobs = []
        generator = QGImages(self.requests, self.responses,
                             1, self.response_blobs)
        querier = ParallelQuery(db)
        querier.query(generator, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert len(self.response_blobs) == len(self.responses)
        for i, resp in enumerate(self.responses):
            for key in resp[0]:
                assert resp[0][key]["returned"] == i + 1
                assert resp[0][key]["status"] == 0
                assert len(self.response_blobs[i]) == resp[0][key]["returned"]

    @pytest.mark.parametrize("max_commands", range(1, 7, 3))
    def test_set_response(self, db, utils, max_commands, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s, r,
                            b: set_query_mocker_factory(5.5)(s, r, b))
        self.cleanDB()
        self.requests = {}
        self.responses = {}
        self.response_blobs = {}
        generator = QGSetPersonAndImages(self.requests, self.responses,
                                         max_commands, self.response_blobs)
        querier = ParallelQuerySet(db)
        querier.query(generator, batchsize=99,
                      numthreads=31,
                      stats=True)

        assert querier.error_counter == 0
        for i, set_key in enumerate(self.responses):
            if set_key in self.response_blobs:
                expected_blobs = 0
                # each resp is the query response back from that query
                for j, resp in enumerate(self.responses[set_key]):
                    for part in resp:
                        key = list(part.keys())[0]
                        expected_status = 0
                        if key == "AddEntity" and j % 2 == 1:
                            expected_status = 2
                        assert part[key]["status"] == expected_status
                        if key == "AddImage":
                            assert (self.response_blobs[set_key][j]
                                    == j * 2)
                            expected_blobs = expected_blobs + 1
                        else:
                            assert part[key]["returned"] == j + 1
                assert len(self.response_blobs[set_key]) == expected_blobs

    @pytest.mark.parametrize("cpq", range(1, 3))
    def test_varying_response_w_index(self, db, utils, cpq):
        self.cleanDB()
        persons = EntityWithResponseDataCSV(
            "./input/persons.adb.csv", self.requests, self.responses)
        loader = ParallelLoader(db)
        loader.ingest(persons, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert loader.error_counter == 0

        self.requests = {}
        self.responses = {}
        self.response_blobs = {}
        generator = QGPersonsIndex(self.requests, self.responses,
                                   cpq, self.response_blobs)
        querier = ParallelQuery(db)
        querier.query(generator, batchsize=100,
                      numthreads=1,
                      stats=True)
        # req and resp are dict mapping index number to query.
        for reqk, respk in zip(self.requests.keys(), self.responses.keys()):
            req = self.requests[reqk]
            age_group = req[0]["FindEntity"]["constraints"]["age"][1]
            # in QGPersonIndex, age_group[1] is created by taking index
            # and multiplying it by cqp and 10.
            assert reqk == (age_group // cpq // 10)

    def test_bad_responsehandler(self, db, utils, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s,
                            r, b: mock_query(s, r, b))
        self.requests = []
        self.responses = []
        self.response_blobs = []
        generator = QGPersonsBadHandler(self.requests, self.responses,
                                        1, self.response_blobs)
        querier = ParallelQuery(db)
        querier.query(generator, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert querier.error_counter != 0



import json
from enum import Enum
import sys
import traceback

import typer
from typing_extensions import Annotated

from aperturedb.cli.console import console

from aperturedb.Connector import Connector
from aperturedb.Images import Images
from aperturedb.ParallelQuery import execute_batch
from aperturedb.Utils import create_connector
import logging

logger = logging.getLogger(__file__)

FUSE_AVAIALBLE = False


def load_fuse():
    global FUSE_AVAIALBLE
    try:
        from aperturedb.cli.mount_coco import mount_images_from_aperturedb
        FUSE_AVAIALBLE = True
    except ImportError as e:
        logger.warning(
            "fuse not found for this env. This is not critical for adb to continue.")


app = typer.Typer(callback=load_fuse)


class OutputTypes(str, Enum):
    STDOUT = "stdout"
    MOUNT_COCO = "mount_coco"


def dump_to_stdout(db: Connector, transaction: dict, **kwargs):
    result, response, blobs = execute_batch(
        db=db,
        q=transaction,
        blobs=[])
    console.log(result)
    console.log(response)
    for i, blob in enumerate(blobs):
        console.log(f"len(blob[{i}]) = {len(blob[i])}")


def mount_as_coco_ds(db: Connector, transaction: dict, **kwargs):
    result, response, blobs = execute_batch(
        db=db,
        q=transaction,
        blobs=[])
    if result == 0:
        image_entities = []
        for i, cr in enumerate(response):
            if "FindImage" in cr:
                if "entities" in cr["FindImage"]:
                    image_entities.extend(cr["FindImage"]["entities"])
                else:
                    console.log(f"No entities found in FindImage {i} response")
        try:
            from aperturedb.cli.mount_coco import mount_images_from_aperturedb
            images = Images(db, response=image_entities)
            console.log(f"Found {len(images)} images")
            mount_images_from_aperturedb(images)
        except Exception as e:
            console.log(traceback.format_exc())
    else:
        console.log(response)


@app.command()
def from_json_file(
    filepath: Annotated[str, typer.Argument(help="Path to query in json format")],
    output_type: Annotated[OutputTypes, typer.Option(
        help="Type of output")] = "stdout",
    output_path: Annotated[str,  typer.Option(
        help="Path to output (only for mount as output)")] = None
):
    db = create_connector()

    output_types = {
        OutputTypes.STDOUT: dump_to_stdout
    }
    global FUSE_AVAIALBLE
    if FUSE_AVAIALBLE:
        output_types[OutputTypes.MOUNT_COCO] = mount_as_coco_ds

    with open(filepath) as inputstream:
        transaction = json.loads(inputstream.read())
        old_argv = sys.argv[1:]
        sys.argv[1:] = [output_path]
        output_types[output_type](db, transaction, output_path=output_path)
        sys.argv[1:] = old_argv



import logging

import numpy as np

from aperturedb.Entities import Entities
from aperturedb.ParallelQuery import execute_batch

logger = logging.getLogger(__name__)


class Descriptors(Entities):
    """
    Python wrapper for ApertureDB Descriptors API.
    """

    db_object = "_Descriptor"

    def __init__(self, db):
        super().__init__(db)

    def find_similar(
        self,
        set: str,
        vector,
        k_neighbors: int,
        constraints=None,
        distances: bool = False,
        blobs: bool = False,
        results={"all_properties": True},
    ):
        """
        Find similar descriptor sets to the input descriptor set.

        Args:
            set (str): Descriptor set name.
            vector (list): Input descriptor set vector.
            k_neighbors (int): Number of neighbors to return.
            distances (bool): Return similarity metric values.
            blobs (bool): Return vectors of the neighbors.
            results (dict): Dictionary with the results format.
                Defaults to all properties.

        Returns:
            results: Response from the server.
        """

        command = {
            "FindDescriptor": {
                "set": set,
                "distances": distances,
                "blobs": blobs,
                "results": results,
                "k_neighbors": k_neighbors,
            }
        }

        if constraints is not None:
            command["FindDescriptor"]["constraints"] = constraints.constraints

        query = [command]
        blobs_in = [np.array(vector, dtype=np.float32).tobytes()]
        _, response, blobs_out = execute_batch(query, blobs_in, self.db)

        self.response = response[0]["FindDescriptor"].get("entities", [])

        if blobs:
            for i, entity in enumerate(self.response):
                entity["vector"] = np.frombuffer(
                    blobs_out[i], dtype=np.float32)

    def _descriptorset_metric(self, set: str):
        """Find default metric for descriptor set"""
        command = {"FindDescriptorSet": {"with_name": set, "metrics": True}}
        query = [command]
        response, _ = self.db.query(query)
        logger.debug(response)
        assert self.db.last_query_ok(), response
        return response[0]["FindDescriptorSet"]["entities"][0]["_metrics"][0]

    def _vector_similarity(self, v1, v2):
        """Find similarity between two vectors using the metric of the descriptor set."""
        if self.metric == "CS":
            return np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
        elif self.metric == "L2":
            # negate to turn distance into similarity
            return -np.linalg.norm(v1 - v2)
        elif self.metric == "IP":
            return np.dot(v1, v2)
        else:
            raise ValueError("Unknown metric: %s" % self.metric)

    def find_similar_mmr(
        self,
        set: str,
        vector,
        k_neighbors: int,
        fetch_k: int,
        lambda_mult: float = 0.5,
        **kwargs,
    ):
        """
        As find_similar, but using the MMR algorithm to diversify the results.

        Args:

            set (str): Descriptor set name.
            vector (list): Input descriptor set vector.
            k_neighbors (int): Number of results to return.
            fetch_k (int): Number of neighbors to fetch from the database.
            lambda_mult (float): Lambda multiplier for the MMR algorithm.
                Defaults to 0.5.  1.0 means no diversity.
        """
        self.metric = self._descriptorset_metric(set)
        vector = np.array(vector, dtype=np.float32)

        kwargs["blobs"] = True  # force vector return
        self.find_similar(set, vector, fetch_k, **kwargs)

        # MMR algorithm
        # Calculate similarity between query and all documents
        query_similarity = [self._vector_similarity(
            vector, d["vector"]) for d in self]
        # Calculate similarity between all pairs of documents
        document_similarity = {}
        for i, d in enumerate(self):
            for j, d2 in enumerate(self[i + 1:], i + 1):
                similarity = self._vector_similarity(d["vector"], d2["vector"])
                document_similarity[(i, j)] = similarity
                document_similarity[(j, i)] = similarity

        # We just gather indexes here, not the actual entities
        selected = []
        unselected = list(range(len(self)))

        while len(selected) < k_neighbors and unselected:
            if not selected:
                selected.append(0)
                unselected.remove(0)
            else:
                selected_unselected_similarity = np.array(
                    [
                        [document_similarity[(i, j)] for j in unselected]
                        for i in selected
                    ]
                )
                worst_similarity = np.max(
                    selected_unselected_similarity, axis=0)
                relevance_scores = np.array(
                    [query_similarity[i] for i in unselected])
                scores = (
                    1 - lambda_mult
                ) * worst_similarity + lambda_mult * relevance_scores
                max_index = unselected[np.argmax(scores)]
                selected.append(max_index)
                unselected.remove(max_index)
        logger.info("Selected indexes: %s; unselected %s",
                    selected, unselected)
        self.response = [self[i] for i in selected]
